@@ -69,24 +69,42 @@
 #include <ctype.h>
 
 static krb5_error_code
-find_alternate_tgs(krb5_kdc_req *,krb5_db_entry **);
+find_alternate_tgs(kdc_realm_t *, krb5_principal, krb5_db_entry **,
+                   const char**);
 
 static krb5_error_code
 prepare_error_tgs(struct kdc_request_state *, krb5_kdc_req *,krb5_ticket *,int,
                   krb5_principal,krb5_data **,const char *, krb5_pa_data **);
 
+static krb5_error_code
+decrypt_2ndtkt(kdc_realm_t *, krb5_kdc_req *, krb5_flags, krb5_db_entry **,
+               const char **);
+
+static krb5_error_code
+gen_session_key(kdc_realm_t *, krb5_kdc_req *, krb5_db_entry *,
+                krb5_keyblock *, const char **);
+
 static krb5_int32
-prep_reprocess_req(krb5_kdc_req *,krb5_principal *);
+find_referral_tgs(kdc_realm_t *, krb5_kdc_req *, krb5_principal *);
+
+static krb5_error_code
+db_get_svc_princ(krb5_context, krb5_principal, krb5_flags,
+                 krb5_db_entry **, const char **);
+
+static krb5_error_code
+search_sprinc(kdc_realm_t *, krb5_kdc_req *, krb5_flags,
+              krb5_db_entry **, const char **);
 
 /*ARGSUSED*/
 krb5_error_code
-process_tgs_req(krb5_data *pkt, const krb5_fulladdr *from,
-                krb5_data **response)
+process_tgs_req(struct server_handle *handle, krb5_data *pkt,
+                const krb5_fulladdr *from, krb5_data **response)
 {
     krb5_keyblock * subkey = 0;
     krb5_keyblock * tgskey = 0;
     krb5_kdc_req *request = 0;
     krb5_db_entry *server = NULL;
+    krb5_db_entry *stkt_server = NULL;
     krb5_kdc_rep reply;
     krb5_enc_kdc_rep_part reply_encpart;
     krb5_ticket ticket_reply, *header_ticket = 0;
@@ -101,12 +119,9 @@ process_tgs_req(krb5_data *pkt, const krb5_fulladdr *from,
     krb5_timestamp rtime;
     krb5_keyblock *reply_key = NULL;
     krb5_key_data  *server_key;
-    char *cname = 0, *sname = 0, *altcname = 0;
+    krb5_principal cprinc = NULL, sprinc = NULL, altcprinc = NULL;
     krb5_last_req_entry *nolrarray[2], nolrentry;
-    krb5_enctype useenctype;
-    int errcode, errcode2;
-    register int i;
-    int firstpass = 1;
+    int errcode;
     const char        *status = 0;
     krb5_enc_tkt_part *header_enc_tkt = NULL; /* TGT */
     krb5_enc_tkt_part *subject_tkt = NULL; /* TGT or evidence ticket */
@@ -114,16 +129,14 @@ process_tgs_req(krb5_data *pkt, const krb5_fulladdr *from,
     krb5_pa_s4u_x509_user *s4u_x509_user = NULL; /* protocol transition request */
     krb5_authdata **kdc_issued_auth_data = NULL; /* auth data issued by KDC */
     unsigned int c_flags = 0, s_flags = 0;       /* client/server KDB flags */
-    char *s4u_name = NULL;
-    krb5_boolean is_referral, db_ref_done = FALSE;
+    krb5_boolean is_referral;
     const char *emsg = NULL;
-    krb5_data *tgs_1 =NULL, *server_1 = NULL;
-    krb5_principal krbtgt_princ;
     krb5_kvno ticket_kvno = 0;
     struct kdc_request_state *state = NULL;
     krb5_pa_data *pa_tgs_req; /*points into request*/
     krb5_data scratch;
     krb5_pa_data **e_data = NULL;
+    kdc_realm_t *kdc_active_realm = NULL;
 
     reply.padata = 0; /* For cleanup handler */
     reply_encpart.enc_padata = 0;
@@ -135,28 +148,28 @@ process_tgs_req(krb5_data *pkt, const krb5_fulladdr *from,
     if (retval)
         return retval;
     if (request->msg_type != KRB5_TGS_REQ) {
-        krb5_free_kdc_req(kdc_context, request);
+        krb5_free_kdc_req(handle->kdc_err_context, request);
         return KRB5_BADMSGTYPE;
     }
 
     /*
      * setup_server_realm() sets up the global realm-specific data pointer.
      */
-    if ((retval = setup_server_realm(request->server))) {
-        krb5_free_kdc_req(kdc_context, request);
-        return retval;
+    kdc_active_realm = setup_server_realm(handle, request->server);
+    if (kdc_active_realm == NULL) {
+        krb5_free_kdc_req(handle->kdc_err_context, request);
+        return KRB5KDC_ERR_WRONG_REALM;
     }
-    errcode = kdc_process_tgs_req(request, from, pkt, &header_ticket,
+    errcode = kdc_make_rstate(kdc_active_realm, &state);
+    if (errcode !=0) {
+        krb5_free_kdc_req(handle->kdc_err_context, request);
+        return errcode;
+    }
+    errcode = kdc_process_tgs_req(kdc_active_realm,
+                                  request, from, pkt, &header_ticket,
                                   &krbtgt, &tgskey, &subkey, &pa_tgs_req);
-    if (header_ticket && header_ticket->enc_part2 &&
-        (errcode2 = krb5_unparse_name(kdc_context,
-                                      header_ticket->enc_part2->client,
-                                      &cname))) {
-        status = "UNPARSING CLIENT";
-        errcode = errcode2;
-        goto cleanup;
-    }
-    limit_string(cname);
+    if (header_ticket && header_ticket->enc_part2)
+        cprinc = header_ticket->enc_part2->client;
 
     if (errcode) {
         status = "PROCESS_TGS";
@@ -168,11 +181,6 @@ process_tgs_req(krb5_data *pkt, const krb5_fulladdr *from,
         status="UNEXPECTED NULL in header_ticket";
         goto cleanup;
     }
-    errcode = kdc_make_rstate(&state);
-    if (errcode !=0) {
-        status = "making state";
-        goto cleanup;
-    }
     scratch.length = pa_tgs_req->length;
     scratch.data = (char *) pa_tgs_req->contents;
     errcode = kdc_find_fast(&request, &scratch, subkey,
@@ -205,65 +213,18 @@ process_tgs_req(krb5_data *pkt, const krb5_fulladdr *from,
         setflag(s_flags, KRB5_KDB_FLAG_CANONICALIZE);
     }
 
-    db_ref_done = FALSE;
-ref_tgt_again:
-    if ((errcode = krb5_unparse_name(kdc_context, request->server, &sname))) {
-        status = "UNPARSING SERVER";
+    errcode = search_sprinc(kdc_active_realm, request, s_flags, &server,
+                            &status);
+    if (errcode != 0)
         goto cleanup;
-    }
-    limit_string(sname);
-
-    errcode = krb5_db_get_principal(kdc_context, request->server,
-                                    s_flags, &server);
-    if (errcode && errcode != KRB5_KDB_NOENTRY) {
-        status = "LOOKING_UP_SERVER";
-        goto cleanup;
-    }
-tgt_again:
-    if (errcode == KRB5_KDB_NOENTRY) {
-        /*
-         * might be a request for a TGT for some other realm; we
-         * should do our best to find such a TGS in this db
-         */
-        if (firstpass ) {
-
-            if ( krb5_is_tgs_principal(request->server) == TRUE) {
-                /* Principal is a name of krb ticket service */
-                if (krb5_princ_size(kdc_context, request->server) == 2) {
-
-                    server_1 = krb5_princ_component(kdc_context,
-                                                    request->server, 1);
-                    tgs_1 = krb5_princ_component(kdc_context, tgs_server, 1);
-
-                    if (!tgs_1 || !data_eq(*server_1, *tgs_1)) {
-                        errcode = find_alternate_tgs(request, &server);
-                        firstpass = 0;
-                        if (errcode == 0)
-                            goto tgt_again;
-                    }
-                }
-                status = "UNKNOWN_SERVER";
-                errcode = KRB5KDC_ERR_S_PRINCIPAL_UNKNOWN;
-                goto cleanup;
-
-            } else if ( db_ref_done == FALSE) {
-                retval = prep_reprocess_req(request, &krbtgt_princ);
-                if (!retval) {
-                    krb5_free_principal(kdc_context, request->server);
-                    retval = krb5_copy_principal(kdc_context, krbtgt_princ,
-                                                 &(request->server));
-                    if (!retval) {
-                        db_ref_done = TRUE;
-                        if (sname != NULL)
-                            free(sname);
-                        goto ref_tgt_again;
-                    }
-                }
-            }
-        }
-
-        status = "UNKNOWN_SERVER";
-        errcode = KRB5KDC_ERR_S_PRINCIPAL_UNKNOWN;
+    sprinc = server->princ;
+    /* XXX until nothing depends on request being mutated */
+    krb5_free_principal(kdc_context, request->server);
+    request->server = NULL;
+    errcode = krb5_copy_principal(kdc_context, server->princ,
+                                  &request->server);
+    if (errcode != 0) {
+        status = "COPYING RESOLVED SERVER";
         goto cleanup;
     }
 
@@ -272,7 +233,8 @@ tgt_again:
         goto cleanup;
     }
 
-    if ((retval = validate_tgs_request(request, *server, header_ticket,
+    if ((retval = validate_tgs_request(kdc_active_realm,
+                                       request, *server, header_ticket,
                                        kdc_time, &status, &e_data))) {
         if (!status)
             status = "UNKNOWN_REASON";
@@ -280,14 +242,14 @@ tgt_again:
         goto cleanup;
     }
 
-    if (!is_local_principal(header_enc_tkt->client))
+    if (!is_local_principal(kdc_active_realm, header_enc_tkt->client))
         setflag(c_flags, KRB5_KDB_FLAG_CROSS_REALM);
 
     is_referral = krb5_is_tgs_principal(server->princ) &&
         !krb5_principal_compare(kdc_context, tgs_server, server->princ);
 
     /* Check for protocol transition */
-    errcode = kdc_process_s4u2self_req(kdc_context,
+    errcode = kdc_process_s4u2self_req(kdc_active_realm,
                                        request,
                                        header_enc_tkt->client,
                                        server,
@@ -302,105 +264,40 @@ tgt_again:
     if (s4u_x509_user != NULL)
         setflag(c_flags, KRB5_KDB_FLAG_PROTOCOL_TRANSITION);
 
-    /*
-     * We pick the session keytype here....
-     *
-     * Some special care needs to be taken in the user-to-user
-     * case, since we don't know what keytypes the application server
-     * which is doing user-to-user authentication can support.  We
-     * know that it at least must be able to support the encryption
-     * type of the session key in the TGT, since otherwise it won't be
-     * able to decrypt the U2U ticket!  So we use that in preference
-     * to anything else.
-     */
-    useenctype = 0;
-    if (isflagset(request->kdc_options, KDC_OPT_ENC_TKT_IN_SKEY |
-                  KDC_OPT_CNAME_IN_ADDL_TKT)) {
-        krb5_keyblock  * st_sealing_key;
-        krb5_kvno        st_srv_kvno;
-        krb5_enctype     etype;
-        krb5_db_entry    *st_client;
-
-        /*
-         * Get the key for the second ticket, and decrypt it.
-         */
-        if ((errcode = kdc_get_server_key(request->second_ticket[st_idx],
-                                          c_flags,
-                                          TRUE, /* match_enctype */
-                                          &st_client,
-                                          &st_sealing_key,
-                                          &st_srv_kvno))) {
-            status = "2ND_TKT_SERVER";
-            goto cleanup;
-        }
-        errcode = krb5_decrypt_tkt_part(kdc_context, st_sealing_key,
-                                        request->second_ticket[st_idx]);
-        krb5_free_keyblock(kdc_context, st_sealing_key);
-        if (errcode) {
-            status = "2ND_TKT_DECRYPT";
-            krb5_db_free_principal(kdc_context, st_client);
-            goto cleanup;
-        }
-
-        etype = request->second_ticket[st_idx]->enc_part2->session->enctype;
-        if (!krb5_c_valid_enctype(etype)) {
-            status = "BAD_ETYPE_IN_2ND_TKT";
-            errcode = KRB5KDC_ERR_ETYPE_NOSUPP;
-            krb5_db_free_principal(kdc_context, st_client);
-            goto cleanup;
-        }
-
-        for (i = 0; i < request->nktypes; i++) {
-            if (request->ktype[i] == etype) {
-                useenctype = etype;
-                break;
-            }
-        }
-
-        if (isflagset(request->kdc_options, KDC_OPT_CNAME_IN_ADDL_TKT)) {
-            /* Do constrained delegation protocol and authorization checks */
-            errcode = kdc_process_s4u2proxy_req(kdc_context,
-                                                request,
-                                                request->second_ticket[st_idx]->enc_part2,
-                                                st_client,
-                                                header_ticket->enc_part2->client,
-                                                request->server,
-                                                &status);
-            if (errcode)
-                goto cleanup;
-
-            setflag(c_flags, KRB5_KDB_FLAG_CONSTRAINED_DELEGATION);
-
-            assert(krb5_is_tgs_principal(header_ticket->server));
-
-            assert(client == NULL); /* assured by kdc_process_s4u2self_req() */
-            client = st_client;
-        } else {
-            /* "client" is not used for user2user */
-            krb5_db_free_principal(kdc_context, st_client);
-        }
-    }
-
-    /*
-     * Select the keytype for the ticket session key.
-     */
-    if ((useenctype == 0) &&
-        (useenctype = select_session_keytype(kdc_context, server,
-                                             request->nktypes,
-                                             request->ktype)) == 0) {
-        /* unsupported ktype */
-        status = "BAD_ENCRYPTION_TYPE";
-        errcode = KRB5KDC_ERR_ETYPE_NOSUPP;
+    errcode = decrypt_2ndtkt(kdc_active_realm, request, c_flags,
+                             &stkt_server, &status);
+    if (errcode)
         goto cleanup;
-    }
 
-    errcode = krb5_c_make_random_key(kdc_context, useenctype, &session_key);
+    if (isflagset(request->kdc_options, KDC_OPT_CNAME_IN_ADDL_TKT)) {
+        /* Do constrained delegation protocol and authorization checks */
+        errcode = kdc_process_s4u2proxy_req(kdc_active_realm,
+                                            request,
+                                            request->second_ticket[st_idx]->enc_part2,
+                                            stkt_server,
+                                            header_ticket->enc_part2->client,
+                                            request->server,
+                                            &status);
+        if (errcode)
+            goto cleanup;
 
-    if (errcode) {
-        /* random key failed */
-        status = "RANDOM_KEY_FAILED";
+        setflag(c_flags, KRB5_KDB_FLAG_CONSTRAINED_DELEGATION);
+
+        assert(krb5_is_tgs_principal(header_ticket->server));
+
+        assert(client == NULL); /* assured by kdc_process_s4u2self_req() */
+        client = stkt_server;
+        stkt_server = NULL;
+    } else if (request->kdc_options & KDC_OPT_ENC_TKT_IN_SKEY) {
+        krb5_db_free_principal(kdc_context, stkt_server);
+        stkt_server = NULL;
+    } else
+        assert(stkt_server == NULL);
+
+    errcode = gen_session_key(kdc_active_realm, request, server, &session_key,
+                              &status);
+    if (errcode)
         goto cleanup;
-    }
 
     /*
      * subject_tkt will refer to the evidence ticket (for constrained
@@ -542,7 +439,7 @@ tgt_again:
         /* not a renew request */
         enc_tkt_reply.times.starttime = kdc_time;
 
-        kdc_get_ticket_endtime(kdc_context, enc_tkt_reply.times.starttime,
+        kdc_get_ticket_endtime(kdc_active_realm, enc_tkt_reply.times.starttime,
                                header_enc_tkt->times.endtime, request->till,
                                client, server, &enc_tkt_reply.times.endtime);
 
@@ -591,19 +488,12 @@ tgt_again:
         enc_tkt_reply.times.starttime = 0;
 
     if (isflagset(c_flags, KRB5_KDB_FLAG_PROTOCOL_TRANSITION)) {
-        errcode = krb5_unparse_name(kdc_context, s4u_x509_user->user_id.user,
-                                    &s4u_name);
+        altcprinc = s4u_x509_user->user_id.user;
     } else if (isflagset(c_flags, KRB5_KDB_FLAG_CONSTRAINED_DELEGATION)) {
-        errcode = krb5_unparse_name(kdc_context, subject_tkt->client,
-                                    &s4u_name);
+        altcprinc = subject_tkt->client;
     } else {
-        errcode = 0;
+        altcprinc = NULL;
     }
-    if (errcode) {
-        status = "UNPARSING S4U CLIENT";
-        goto cleanup;
-    }
-
     if (isflagset(request->kdc_options, KDC_OPT_ENC_TKT_IN_SKEY)) {
         krb5_enc_tkt_part *t2enc = request->second_ticket[st_idx]->enc_part2;
         encrypting_key = *(t2enc->session);
@@ -699,8 +589,9 @@ tgt_again:
      * listed).
      */
     /* realm compare is like strcmp, but knows how to deal with these args */
-    if (realm_compare(header_ticket->server, tgs_server) ||
-        realm_compare(header_ticket->server, enc_tkt_reply.client)) {
+    if (krb5_realm_compare(kdc_context, header_ticket->server, tgs_server) ||
+        krb5_realm_compare(kdc_context, header_ticket->server,
+                           enc_tkt_reply.client)) {
         /* tgt issued by local realm or issued by realm of client */
         enc_tkt_reply.transited = header_enc_tkt->transited;
     } else {
@@ -738,35 +629,15 @@ tgt_again:
         }
     }
     if (!isflagset (request->kdc_options, KDC_OPT_DISABLE_TRANSITED_CHECK)) {
-        unsigned int tlen;
-        char *tdots;
-
-        errcode = kdc_check_transited_list (kdc_context,
+        errcode = kdc_check_transited_list (kdc_active_realm,
                                             &enc_tkt_reply.transited.tr_contents,
                                             krb5_princ_realm (kdc_context, header_enc_tkt->client),
                                             krb5_princ_realm (kdc_context, request->server));
-        tlen = enc_tkt_reply.transited.tr_contents.length;
-        tdots = tlen > 125 ? "..." : "";
-        tlen = tlen > 125 ? 125 : tlen;
-
         if (errcode == 0) {
             setflag (enc_tkt_reply.flags, TKT_FLG_TRANSIT_POLICY_CHECKED);
-        } else if (errcode == KRB5KRB_AP_ERR_ILL_CR_TKT)
-            krb5_klog_syslog(LOG_INFO, _("bad realm transit path from '%s' "
-                                         "to '%s' via '%.*s%s'"),
-                             cname ? cname : "<unknown client>",
-                             sname ? sname : "<unknown server>", tlen,
-                             enc_tkt_reply.transited.tr_contents.data, tdots);
-        else {
-            emsg = krb5_get_error_message(kdc_context, errcode);
-            krb5_klog_syslog(LOG_ERR, _("unexpected error checking transit "
-                                        "from '%s' to '%s' via '%.*s%s': %s"),
-                             cname ? cname : "<unknown client>",
-                             sname ? sname : "<unknown server>", tlen,
-                             enc_tkt_reply.transited.tr_contents.data, tdots,
-                             emsg);
-            krb5_free_error_message(kdc_context, emsg);
-            emsg = NULL;
+        } else {
+            log_tgs_badtrans(kdc_context, cprinc, sprinc,
+                             &enc_tkt_reply.transited.tr_contents, errcode);
         }
     } else
         krb5_klog_syslog(LOG_INFO, _("not checking transit path"));
@@ -793,11 +664,7 @@ tgt_again:
         krb5_enc_tkt_part *t2enc = request->second_ticket[st_idx]->enc_part2;
         krb5_principal client2 = t2enc->client;
         if (!krb5_principal_compare(kdc_context, request->server, client2)) {
-            if ((errcode = krb5_unparse_name(kdc_context, client2, &altcname)))
-                altcname = 0;
-            if (altcname != NULL)
-                limit_string(altcname);
-
+            altcprinc = client2;
             errcode = KRB5KDC_ERR_SERVER_NOMATCH;
             status = "2ND_TKT_MISMATCH";
             goto cleanup;
@@ -822,7 +689,8 @@ tgt_again:
     /* Start assembling the response */
     reply.msg_type = KRB5_TGS_REP;
     if (isflagset(c_flags, KRB5_KDB_FLAG_PROTOCOL_TRANSITION) &&
-        find_pa_data(request->padata, KRB5_PADATA_S4U_X509_USER) != NULL) {
+        krb5int_find_pa_data(kdc_context, request->padata,
+                             KRB5_PADATA_S4U_X509_USER) != NULL) {
         errcode = kdc_make_s4u2self_rep(kdc_context,
                                         subkey,
                                         header_ticket->enc_part2->session,
@@ -910,8 +778,9 @@ cleanup:
         krb5_free_keyblock(kdc_context, reply_key);
     if (errcode)
         emsg = krb5_get_error_message (kdc_context, errcode);
-    log_tgs_req(from, request, &reply, cname, sname, altcname, authtime,
-                c_flags, s4u_name, status, errcode, emsg);
+    log_tgs_req(kdc_context, from, request, &reply, cprinc,
+                sprinc, altcprinc, authtime,
+                c_flags, status, errcode, emsg);
     if (errcode) {
         krb5_free_error_message (kdc_context, emsg);
         emsg = NULL;
@@ -942,10 +811,6 @@ cleanup:
         krb5_free_kdc_req(kdc_context, request);
     if (state)
         kdc_free_rstate(state);
-    if (cname != NULL)
-        free(cname);
-    if (sname != NULL)
-        free(sname);
     krb5_db_free_principal(kdc_context, server);
     krb5_db_free_principal(kdc_context, krbtgt);
     krb5_db_free_principal(kdc_context, client);
@@ -957,8 +822,6 @@ cleanup:
         krb5_free_pa_s4u_x509_user(kdc_context, s4u_x509_user);
     if (kdc_issued_auth_data != NULL)
         krb5_free_authdata(kdc_context, kdc_issued_auth_data);
-    if (s4u_name != NULL)
-        free(s4u_name);
     if (subkey != NULL)
         krb5_free_keyblock(kdc_context, subkey);
     if (tgskey != NULL)
@@ -984,6 +847,7 @@ prepare_error_tgs (struct kdc_request_state *state,
     krb5_error errpkt;
     krb5_error_code retval = 0;
     krb5_data *scratch, *e_data_asn1 = NULL, *fast_edata = NULL;
+    kdc_realm_t *kdc_active_realm = state->realm_data;
 
     errpkt.ctime = request->nonce;
     errpkt.cusec = 0;
@@ -1041,33 +905,136 @@ prepare_error_tgs (struct kdc_request_state *state,
     return retval;
 }
 
+/* KDC options that require a second ticket */
+#define STKT_OPTIONS (KDC_OPT_CNAME_IN_ADDL_TKT | KDC_OPT_ENC_TKT_IN_SKEY)
+/*
+ * Get the key for the second ticket, if any, and decrypt it.
+ */
+static krb5_error_code
+decrypt_2ndtkt(kdc_realm_t *kdc_active_realm, krb5_kdc_req *req,
+               krb5_flags flags, krb5_db_entry **server_out,
+               const char **status)
+{
+    krb5_error_code retval;
+    krb5_db_entry *server;
+    krb5_keyblock *key;
+    krb5_kvno kvno;
+    krb5_ticket *stkt;
+
+    if (!(req->kdc_options & STKT_OPTIONS))
+        return 0;
+
+    stkt = req->second_ticket[0];
+    retval = kdc_get_server_key(kdc_context, stkt,
+                                flags,
+                                TRUE, /* match_enctype */
+                                &server,
+                                &key,
+                                &kvno);
+    if (retval != 0) {
+        *status = "2ND_TKT_SERVER";
+        goto cleanup;
+    }
+    retval = krb5_decrypt_tkt_part(kdc_context, key,
+                                   req->second_ticket[0]);
+    krb5_free_keyblock(kdc_context, key);
+    if (retval != 0) {
+        *status = "2ND_TKT_DECRYPT";
+        goto cleanup;
+    }
+    *server_out = server;
+cleanup:
+    return retval;
+}
+
+static krb5_error_code
+get_2ndtkt_enctype(kdc_realm_t *kdc_active_realm, krb5_kdc_req *req,
+                   krb5_enctype *useenctype, const char **status)
+{
+    krb5_enctype etype;
+    krb5_ticket *stkt = req->second_ticket[0];
+    int i;
+
+    etype = stkt->enc_part2->session->enctype;
+    if (!krb5_c_valid_enctype(etype)) {
+        *status = "BAD_ETYPE_IN_2ND_TKT";
+        return KRB5KDC_ERR_ETYPE_NOSUPP;
+    }
+    for (i = 0; i < req->nktypes; i++) {
+        if (req->ktype[i] == etype) {
+            *useenctype = etype;
+            break;
+        }
+    }
+    return 0;
+}
+
+static krb5_error_code
+gen_session_key(kdc_realm_t *kdc_active_realm, krb5_kdc_req *req,
+                krb5_db_entry *server, krb5_keyblock *skey,
+                const char **status)
+{
+    krb5_error_code retval;
+    krb5_enctype useenctype = 0;
+
+    /*
+     * Some special care needs to be taken in the user-to-user
+     * case, since we don't know what keytypes the application server
+     * which is doing user-to-user authentication can support.  We
+     * know that it at least must be able to support the encryption
+     * type of the session key in the TGT, since otherwise it won't be
+     * able to decrypt the U2U ticket!  So we use that in preference
+     * to anything else.
+     */
+    if (req->kdc_options & KDC_OPT_ENC_TKT_IN_SKEY) {
+        retval = get_2ndtkt_enctype(kdc_active_realm, req, &useenctype,
+                                    status);
+        if (retval != 0)
+            goto cleanup;
+    }
+    if (useenctype == 0) {
+        useenctype = select_session_keytype(kdc_active_realm, server,
+                                            req->nktypes,
+                                            req->ktype);
+    }
+    if (useenctype == 0) {
+        /* unsupported ktype */
+        *status = "BAD_ENCRYPTION_TYPE";
+        retval = KRB5KDC_ERR_ETYPE_NOSUPP;
+        goto cleanup;
+    }
+    retval = krb5_c_make_random_key(kdc_context, useenctype, skey);
+    if (retval != 0) {
+        /* random key failed */
+        *status = "RANDOM_KEY_FAILED";
+        goto cleanup;
+    }
+cleanup:
+    return retval;
+}
+
 /*
  * The request seems to be for a ticket-granting service somewhere else,
  * but we don't have a ticket for the final TGS.  Try to give the requestor
  * some intermediate realm.
  */
 static krb5_error_code
-find_alternate_tgs(krb5_kdc_req *request, krb5_db_entry **server_ptr)
+find_alternate_tgs(kdc_realm_t *kdc_active_realm, krb5_principal princ,
+                   krb5_db_entry **server_ptr, const char **status)
 {
     krb5_error_code retval;
-    krb5_principal *plist = NULL, *pl2, tmpprinc;
+    krb5_principal *plist = NULL, *pl2;
     krb5_data tmp;
     krb5_db_entry *server = NULL;
 
     *server_ptr = NULL;
-
-    /*
-     * Call to krb5_princ_component is normally not safe but is so
-     * here only because find_alternate_tgs() is only called from
-     * somewhere that has already checked the number of components in
-     * the principal.
-     */
+    assert(is_cross_tgs_principal(princ));
     if ((retval = krb5_walk_realm_tree(kdc_context,
-                                       krb5_princ_realm(kdc_context, request->server),
-                                       krb5_princ_component(kdc_context, request->server, 1),
-                                       &plist, KRB5_REALM_BRANCH_CHAR)))
-        return retval;
-
+                                       krb5_princ_realm(kdc_context, princ),
+                                       krb5_princ_component(kdc_context, princ, 1),
+                                       &plist, KRB5_REALM_BRANCH_CHAR))) {
+        goto cleanup;
+    }
     /* move to the end */
     for (pl2 = plist; *pl2; pl2++);
 
@@ -1077,130 +1044,174 @@ find_alternate_tgs(krb5_kdc_req *request, krb5_db_entry **server_ptr)
         tmp = *krb5_princ_realm(kdc_context, *pl2);
         krb5_princ_set_realm(kdc_context, *pl2,
                              krb5_princ_realm(kdc_context, tgs_server));
-        retval = krb5_db_get_principal(kdc_context, *pl2, 0, &server);
+        retval = db_get_svc_princ(kdc_context, *pl2, 0, &server, status);
         krb5_princ_set_realm(kdc_context, *pl2, &tmp);
         if (retval == KRB5_KDB_NOENTRY)
             continue;
         else if (retval)
             goto cleanup;
 
-        /* Found it. */
-        tmp = *krb5_princ_realm(kdc_context, *pl2);
-        krb5_princ_set_realm(kdc_context, *pl2,
-                             krb5_princ_realm(kdc_context, tgs_server));
-        retval = krb5_copy_principal(kdc_context, *pl2, &tmpprinc);
-        if (retval)
-            goto cleanup;
-        krb5_princ_set_realm(kdc_context, *pl2, &tmp);
-
-        krb5_free_principal(kdc_context, request->server);
-        request->server = tmpprinc;
-        log_tgs_alt_tgt(request->server);
+        log_tgs_alt_tgt(kdc_context, server->princ);
         *server_ptr = server;
         server = NULL;
         goto cleanup;
     }
-    retval = KRB5_KDB_NOENTRY;
-
 cleanup:
+    if (retval == 0 && server_ptr == NULL)
+        retval = KRB5_KDB_NOENTRY;
+    if (retval != 0)
+        *status = "UNKNOWN_SERVER";
+
     krb5_free_realm_tree(kdc_context, plist);
     krb5_db_free_principal(kdc_context, server);
     return retval;
 }
 
-static krb5_int32
-prep_reprocess_req(krb5_kdc_req *request, krb5_principal *krbtgt_princ)
+/*
+ * Check whether the request satisfies the conditions for generating a referral
+ * TGT.  The caller checks whether the hostname component looks like a FQDN.
+ */
+static krb5_boolean
+is_referral_req(kdc_realm_t *kdc_active_realm, krb5_kdc_req *request)
 {
-    krb5_error_code retval = KRB5KRB_AP_ERR_BADMATCH;
-    char **realms, **cpp, *temp_buf=NULL;
-    krb5_data *comp1 = NULL, *comp2 = NULL;
-    char *comp1_str = NULL;
+    krb5_boolean ret = FALSE;
+    char *stype = NULL;
+    char *ref_services = kdc_active_realm->realm_host_based_services;
+    char *nonref_services = kdc_active_realm->realm_no_host_referral;
 
-    /* By now we know that server principal name is unknown.
-     * If CANONICALIZE flag is set in the request
-     * If req is not U2U authn. req
-     * the requested server princ. has exactly two components
-     * either
-     *      the name type is NT-SRV-HST
-     *      or name type is NT-UNKNOWN and
-     *         the 1st component is listed in conf file under host_based_services
-     * the 1st component is not in a list in conf under "no_host_referral"
-     * the 2d component looks like fully-qualified domain name (FQDN)
-     * If all of these conditions are satisfied - try mapping the FQDN and
-     * re-process the request as if client had asked for cross-realm TGT.
-     */
-    if (isflagset(request->kdc_options, KDC_OPT_CANONICALIZE) &&
-        !isflagset(request->kdc_options, KDC_OPT_ENC_TKT_IN_SKEY) &&
-        krb5_princ_size(kdc_context, request->server) == 2) {
+    if (!(request->kdc_options & KDC_OPT_CANONICALIZE))
+        return FALSE;
 
-        comp1 = krb5_princ_component(kdc_context, request->server, 0);
-        comp2 = krb5_princ_component(kdc_context, request->server, 1);
+    if (request->kdc_options & KDC_OPT_ENC_TKT_IN_SKEY)
+        return FALSE;
 
-        comp1_str = calloc(1,comp1->length+1);
-        if (!comp1_str) {
-            retval = ENOMEM;
+    if (krb5_princ_size(kdc_context, request->server) != 2)
+        return FALSE;
+
+    stype = data2string(krb5_princ_component(kdc_context, request->server, 0));
+    if (stype == NULL)
+        return FALSE;
+    switch (krb5_princ_type(kdc_context, request->server)) {
+    case KRB5_NT_UNKNOWN:
+        /* Allow referrals for NT-UNKNOWN principals, if configured. */
+        if (kdc_active_realm->realm_host_based_services != NULL) {
+            if (!krb5_match_config_pattern(ref_services, stype) &&
+                !krb5_match_config_pattern(ref_services, KRB5_CONF_ASTERISK))
+                goto cleanup;
+        } else
             goto cleanup;
+        /* FALLTHROUGH */
+    case KRB5_NT_SRV_HST:
+    case KRB5_NT_SRV_INST:
+        /* Deny referrals for specific service types, if configured. */
+        if (kdc_active_realm->realm_no_host_referral != NULL) {
+            if (krb5_match_config_pattern(nonref_services, stype))
+                goto cleanup;
+            if (krb5_match_config_pattern(nonref_services, KRB5_CONF_ASTERISK))
+                goto cleanup;
         }
-        if (comp1->data != NULL)
-            memcpy(comp1_str, comp1->data, comp1->length);
-
-        if ((krb5_princ_type(kdc_context, request->server) == KRB5_NT_SRV_HST ||
-             krb5_princ_type(kdc_context, request->server) == KRB5_NT_SRV_INST ||
-             (krb5_princ_type(kdc_context, request->server) == KRB5_NT_UNKNOWN &&
-              kdc_active_realm->realm_host_based_services != NULL &&
-              (krb5_match_config_pattern(kdc_active_realm->realm_host_based_services,
-                                         comp1_str) == TRUE ||
-               krb5_match_config_pattern(kdc_active_realm->realm_host_based_services,
-                                         KRB5_CONF_ASTERISK) == TRUE))) &&
-            (kdc_active_realm->realm_no_host_referral == NULL ||
-             (krb5_match_config_pattern(kdc_active_realm->realm_no_host_referral,
-                                        KRB5_CONF_ASTERISK) == FALSE &&
-              krb5_match_config_pattern(kdc_active_realm->realm_no_host_referral,
-                                        comp1_str) == FALSE))) {
-
-            if (memchr(comp2->data, '.', comp2->length) == NULL)
-                goto cleanup;
-            temp_buf = calloc(1, comp2->length+1);
-            if (!temp_buf) {
-                retval = ENOMEM;
-                goto cleanup;
-            }
-            if (comp2->data != NULL)
-                memcpy(temp_buf, comp2->data, comp2->length);
-            retval = krb5int_get_domain_realm_mapping(kdc_context, temp_buf, &realms);
-            free(temp_buf);
-            if (retval) {
-                /* no match found */
-                kdc_err(kdc_context, retval, "unable to find realm of host");
-                goto cleanup;
-            }
-            if (realms == 0) {
-                retval = KRB5KRB_AP_ERR_BADMATCH;
-                goto cleanup;
-            }
-            /* Don't return a referral to the null realm or the service
-             * realm. */
-            if (realms[0] == 0 ||
-                data_eq_string(request->server->realm, realms[0])) {
-                free(realms[0]);
-                free(realms);
-                retval = KRB5KRB_AP_ERR_BADMATCH;
-                goto cleanup;
-            }
-            /* Modify request.
-             * Construct cross-realm tgt :  krbtgt/REMOTE_REALM@LOCAL_REALM
-             * and use it as a principal in this req.
-             */
-            retval = krb5_build_principal(kdc_context, krbtgt_princ,
-                                          (*request->server).realm.length,
-                                          (*request->server).realm.data,
-                                          "krbtgt", realms[0], (char *)0);
-            for (cpp = realms; *cpp; cpp++)
-                free(*cpp);
-        }
+        ret = TRUE;
+        break;
+    default:
+        goto cleanup;
     }
 cleanup:
-    free(comp1_str);
+    free(stype);
+    return ret;
+}
+
+/*
+ * Find a remote realm TGS principal for an unknown host-based service
+ * principal.
+ */
+static krb5_int32
+find_referral_tgs(kdc_realm_t *kdc_active_realm, krb5_kdc_req *request,
+                  krb5_principal *krbtgt_princ)
+{
+    krb5_error_code retval = KRB5KDC_ERR_S_PRINCIPAL_UNKNOWN;
+    char **realms = NULL, *hostname = NULL;
+    krb5_data srealm = request->server->realm;
+
+    if (!is_referral_req(kdc_active_realm, request))
+        goto cleanup;
+
+    hostname = data2string(krb5_princ_component(kdc_context,
+                                                request->server, 1));
+    if (hostname == NULL) {
+        retval = ENOMEM;
+        goto cleanup;
+    }
+    /* If the hostname doesn't contain a '.', it's not a FQDN. */
+    if (strchr(hostname, '.') == NULL)
+        goto cleanup;
+    retval = krb5_get_host_realm(kdc_context, hostname, &realms);
+    if (retval) {
+        /* no match found */
+        kdc_err(kdc_context, retval, "unable to find realm of host");
+        goto cleanup;
+    }
+    /* Don't return a referral to the empty realm or the service realm. */
+    if (realms == NULL || realms[0] == NULL || *realms[0] == '\0' ||
+        data_eq_string(srealm, realms[0])) {
+        retval = KRB5KDC_ERR_S_PRINCIPAL_UNKNOWN;
+        goto cleanup;
+    }
+    retval = krb5_build_principal(kdc_context, krbtgt_princ,
+                                  srealm.length, srealm.data,
+                                  "krbtgt", realms[0], (char *)0);
+cleanup:
+    krb5_free_host_realm(kdc_context, realms);
+    free(hostname);
 
     return retval;
 }
+
+static krb5_error_code
+db_get_svc_princ(krb5_context ctx, krb5_principal princ,
+                 krb5_flags flags, krb5_db_entry **server,
+                 const char **status)
+{
+    krb5_error_code ret;
+
+    ret = krb5_db_get_principal(ctx, princ, flags, server);
+    if (ret == KRB5_KDB_CANTLOCK_DB)
+        ret = KRB5KDC_ERR_SVC_UNAVAILABLE;
+    if (ret != 0) {
+        *status = "LOOKING_UP_SERVER";
+    }
+    return ret;
+}
+
+static krb5_error_code
+search_sprinc(kdc_realm_t *kdc_active_realm, krb5_kdc_req *req,
+              krb5_flags flags, krb5_db_entry **server, const char **status)
+{
+    krb5_error_code ret;
+    krb5_principal princ = req->server;
+    krb5_principal reftgs = NULL;
+
+    ret = db_get_svc_princ(kdc_context, princ, flags, server, status);
+    if (ret == 0 || ret != KRB5_KDB_NOENTRY)
+        goto cleanup;
+
+    if (!is_cross_tgs_principal(req->server)) {
+        ret = find_referral_tgs(kdc_active_realm, req, &reftgs);
+        if (ret != 0)
+            goto cleanup;
+        ret = db_get_svc_princ(kdc_context, reftgs, flags, server, status);
+        if (ret == 0 || ret != KRB5_KDB_NOENTRY)
+            goto cleanup;
+
+        princ = reftgs;
+    }
+    ret = find_alternate_tgs(kdc_active_realm, princ, server, status);
+
+cleanup:
+    if (ret != 0 && ret != KRB5KDC_ERR_SVC_UNAVAILABLE) {
+        ret = KRB5KDC_ERR_S_PRINCIPAL_UNKNOWN;
+        if (*status == NULL)
+            *status = "LOOKING_UP_SERVER";
+    }
+    krb5_free_principal(kdc_context, reftgs);
+    return ret;
+}
